//
//  DataMonitorView.m
//  GlowWorm
//
//  Created by Curtis Jones on 2007.02.08.
//  Copyright 2007 __MyCompanyName__. All rights reserved.
//

#import <DataMonitorView.h>
#import <Common.h>
#import <Logger.h>
#import <GlowWorm.h>
#import <DataMonitorController.h>

#import <stdlib.h>
#import <sys/time.h>
#import <string.h>

char * strcpy(char *, const char *);
void bzero(void *, size_t);

@implementation DataMonitorView

+ (uint64_t)getTimeInMicros
{
	struct timeval now;
	gettimeofday(&now, NULL);
	return ((uint64_t)now.tv_sec * 1000000) + (uint64_t)now.tv_usec;
}

- (id)initWithFrame:(NSRect)frameRect
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	self = [super initWithFrame:frameRect];
	
	if (self) {
		mOffset		= 0;
		mBuffer		= 50;
		mScale		= 100;
		mCount		= 0;
		mRate			= 200000;
		mSamples	= [[NSMutableArray alloc] initWithCapacity:10];
		mTime			= [DataMonitorView getTimeInMicros];
		mFadeAlpha = 0.;
		
		/*
		NSString * imgPath = 
		CFStringCreateWithCString(kCFAllocatorDefault, , kCFStringEncodingMacRoman);
		*/
		
		{
			int i = 0;
			GlowWormStat * stat = [[GlowWormStat alloc] init];
			
			for (i = 0; i < mBuffer; ++i)
				[mSamples addObject:stat];
			
			[stat release];
		}
		
#ifndef STANDALONE
		[NSThread detachNewThreadSelector:@selector(redrawLoop) toTarget:self withObject:nil];
#else
		[NSThread detachNewThreadSelector:@selector(redrawLoop) toTarget:self withObject:nil];
		[NSThread detachNewThreadSelector:@selector(randomDataLoop) toTarget:self withObject:nil];
#endif
	}
	
	return self;
}

- (void)dealloc
{
	mHault = TRUE;
	
	@synchronized (mSamples) {
		
	}
	
	[mSamples release];
	
	[super dealloc];
}

- (void)viewDidMoveToWindow
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	[self addTrackingRect:[self bounds] owner:self userData:nil assumeInside:FALSE];
	
//mEditBtn = ((DataMonitorController *)[self window])->mEditBtn;
//[mEditBtn setTransparent:TRUE];
}

- (void)addSample:(GlowWormStat *)aStat
{
	if (aStat == nil)
		return;
	
	@synchronized (mSamples)
	{
		uint64_t now = [DataMonitorView getTimeInMicros];
		
		int i=0, dif = (mTime == 0 ? 0 : ((now-mTime)/mRate));
		
	//GWLog(@"%s.. mOffset = %d, mBuffer = %d, mTime = %llu, dif = %d", __PRETTY_FUNCTION__, mOffset, mBuffer, mTime, dif);
		
		// if enough time units have passed such that we would
		// need to zero the entire circular list, then just 
		// reset the entry count.
		//
		if (dif >= mBuffer) {
			int i = 0;
			GlowWormStat * stat = [[GlowWormStat alloc] init];
			
			for (i = 0; i < mBuffer; ++i)
				[mSamples replaceObjectAtIndex:i withObject:stat];
			
			[stat release];
		}
		
		// if multiple time units have passed since the last 
		// sample data was received, then fill in the missing
		// samples with copies of the previous sample, thus
		// indicating zero-transfer during that time.
		//
		else if (dif > 1)
			for (i = mOffset; i < mOffset+dif-1; ++i)
//			[mSamples replaceObjectAtIndex:((i+1)%mBuffer) withObject:[mSamples objectAtIndex:i == 0 ? mBuffer-1 : ((i-1)%mBuffer == 0 ? mBuffer - 1 : (i-1))]];
				[mSamples replaceObjectAtIndex:((i+1)%mBuffer) withObject:[mSamples objectAtIndex:i%mBuffer]];
		
//	if (dif != 0 && mCount < mBuffer)
//		mCount++;
		
		mOffset = (mOffset + dif) % mBuffer;
		
		[mSamples replaceObjectAtIndex:mOffset withObject:aStat];
		
		if (mTime == 0 || dif != 0)
			mTime = now;
	}
}

/*
static void
RedBlackRedRampEvaluate (void *info, const float *in, float *out)
{
	out[0] = fabs(1. - in[0] * 2);
	out[1] = out[2] = 0;
	out[3] = 1;
}

static CGFunctionRef
createFunctionForRGB (CGFunctionEvaluateCallback evalFunc)
{
	CGFunctionRef func;
	CGFunctionCallbacks shadingCallbacks;
	
	float domain[2];
	float range[8];
	
	domain[0] = 0;
	domain[1] = 1;
	
	range[0] = 0;
	range[1] = 1;
	range[2] = 0;
	range[3] = 1;
	range[4] = 0;
	range[5] = 1;
	range[6] = 0;
	range[7] = 1;
	
	shadingCallbacks.version = 0;
	shadingCallbacks.evaluate = evalFunc;
	shadingCallbacks.releaseInfo = NULL;
	
	func = CGFunctionCreate(NULL, 1, domain, 4, range, &shadingCallbacks);
	
	return func;
}

#define kGenericRGBProfilePathStr "/System/Library/ColorSync/Profiles/Generic RGB Profile.icc"

CGColorSpaceRef
getTheRGBColorSpace ()
{
	static CGColorSpaceRef genericRGBColorSpace = NULL;
	
	CMProfileRef genericRGBProfile = NULL;
	OSStatus err = noErr;
	CMProfileLocation loc;
	
	loc.locType = cmPathBasedProfile;
	strcpy(loc.u.pathLoc.path, kGenericRGBProfilePathStr);
	err = CMOpenProfile(&genericRGBProfile, &loc);
	
	if (err == noErr) {
		genericRGBColorSpace = CGColorSpaceCreateWithPlatformColorSpace(genericRGBProfile);
		
		if (genericRGBColorSpace == NULL)
			GWLog(@"%s.. B", __PRETTY_FUNCTION__);
		
		CMCloseProfile(genericRGBProfile);
	}
	else
		GWLog(@"%s.. A, %d", __PRETTY_FUNCTION__, err);
	
	return genericRGBColorSpace;
}
*/

- (void)drawRect:(NSRect)rect
{
//GWLog(@"%s.. w=%d, h=%d", __PRETTY_FUNCTION__, (int)[self bounds].size.width, (int)[self bounds].size.height);
	
	if (mHault)
		return;
	
	CGContextRef myContext = [[NSGraphicsContext currentContext] graphicsPort];
	CGPoint begPt, endPt;
	
	int i=0, offs=0;
	uint64_t tmpPt=0, highPt=0;
	uint64_t now = [DataMonitorView getTimeInMicros];
	
	if (myContext == NULL) {
		GWLog(@"%s.. invalid context, ignoring", __PRETTY_FUNCTION__);
		return;
	}
	
	if (mLiveResize)
	{
		float size = 9.;
		float color[] = {1., 1., 1., 1.};
		NSString * hw = [NSString stringWithFormat:@"%d x %d", (int)[self bounds].size.width, (int)[self bounds].size.height];
		
		CGContextSelectFont(myContext, "Lucida Grande", size, kCGEncodingMacRoman);
		CGContextSetFillColor(myContext, color);
		CGContextShowTextAtPoint(myContext, ([self bounds].size.width/2)-20, ([self bounds].size.height/2)-8, (char *)[hw UTF8String], [hw length]);
		
		return;
	}
	
	@synchronized (mSamples)
	{
		// find the high point in terms of transfer-per-sample and 
		// determine whether a scale change is necessary.
		//
		{
			uint64_t prev_bytes_send=0, prev_bytes_recv=0;
			uint64_t prev_pckts_send=0, prev_pckts_recv=0;
			
			for (i=0, offs=mOffset+1; i < mBuffer-1; ++i, ++offs) {
				GlowWormStat * stat = [mSamples objectAtIndex:offs%mBuffer];
				
				if (prev_bytes_send != 0 && (tmpPt = [[stat bytesSend] unsignedLongLongValue]) > prev_bytes_send && (tmpPt = (tmpPt - prev_bytes_send)) > highPt) highPt = tmpPt;
				if (prev_bytes_recv != 0 && (tmpPt = [[stat bytesRecv] unsignedLongLongValue]) > prev_bytes_recv && (tmpPt = (tmpPt - prev_bytes_recv)) > highPt) highPt = tmpPt;
				if (prev_pckts_send != 0 && (tmpPt = [[stat pcktsSend] unsignedLongLongValue]) > prev_pckts_send && (tmpPt = (tmpPt - prev_pckts_send)) > highPt) highPt = tmpPt;
				if (prev_pckts_recv != 0 && (tmpPt = [[stat pcktsRecv] unsignedLongLongValue]) > prev_pckts_recv && (tmpPt = (tmpPt - prev_pckts_recv)) > highPt) highPt = tmpPt;
				
				prev_bytes_send = [[stat bytesSend] unsignedLongLongValue];
				prev_bytes_recv = [[stat bytesRecv] unsignedLongLongValue];
				prev_pckts_send = [[stat pcktsSend] unsignedLongLongValue];
				prev_pckts_recv = [[stat pcktsRecv] unsignedLongLongValue];
			}
			
			if (highPt >= mScale)
				mScale = highPt * 1.2;
			else if (highPt < mScale / 2)
				mScale *= 0.75;
			
		//GWLog(@"%s.. highPt = %llu, scale = %d", __PRETTY_FUNCTION__, highPt, mScale);
		}
		
		// draw each sample using the pre-determined scale
		//
		{
			uint32_t height = (uint32_t)[self bounds].size.height;
			uint32_t width  = (uint32_t)[self bounds].size.width - 1;
			uint32_t doffs  = 1;
			uint32_t dheight = 0;
			uint32_t dwidth = width / (mBuffer-2);
			uint32_t dextra = width - (dwidth * (mBuffer-2));
			uint32_t toff   = (now - mTime) / mRate;
			
			uint64_t prev_bytes_send=0, prev_bytes_recv=0;
			uint64_t prev_pckts_send=0, prev_pckts_recv=0;
			
			prev_bytes_send = [[[mSamples objectAtIndex:(mOffset+1)%mBuffer] bytesSend] unsignedLongLongValue];
			prev_bytes_recv = [[[mSamples objectAtIndex:(mOffset+1)%mBuffer] bytesRecv] unsignedLongLongValue];
			
			for (i=toff, offs=mOffset+2+toff; i < mBuffer-2; ++i, ++offs, doffs+=dwidth) {
				GlowWormStat * stat = [mSamples objectAtIndex:offs%mBuffer];
				
				uint32_t partial = (dextra / ((mBuffer-1) - i));
				
				if (partial * ((mBuffer-1)-i) < dextra)
					partial++;
				
				if (prev_bytes_send != 0 && (tmpPt = [[stat bytesSend] unsignedLongLongValue]) > prev_bytes_send) {
					tmpPt -= prev_bytes_send;
					dheight = (uint32_t)(((float)height / (float)mScale) * (float)tmpPt);
					
					if (dheight < (dwidth/2) + 1)
						dheight = (dwidth / 2) + 1;
					
					begPt.x = (float)doffs;
					begPt.y = 10.;
					endPt.x = (float)doffs + ((float)dwidth-1.) + (float)partial;
					endPt.y = 10.;
					
					CGContextSetRGBFillColor(myContext, 1, 0, 0, .5);
					CGContextFillRect(myContext, CGRectMake(doffs, (dheight - (dwidth/2)), dwidth, (dwidth/2)));
				}
				else {
					CGContextSetRGBFillColor(myContext, 0, 0, 1, .5);
					CGContextFillRect(myContext, CGRectMake((float)doffs, 1, (dwidth-1)+partial, 1));
				}
				
				if (prev_bytes_recv != 0 && (tmpPt = [[stat bytesRecv] unsignedLongLongValue]) > prev_bytes_recv) {
					tmpPt -= prev_bytes_recv;
					dheight = (uint32_t)(((float)height / (float)mScale) * (float)tmpPt);
					
					if (dheight < (dwidth/2) + 1)
						dheight = (dwidth / 2) + 1;
					
					begPt.x = (float)doffs;
					begPt.y = 10.;
					endPt.x = (float)doffs + ((float)dwidth-1.) + (float)partial;
					endPt.y = 10.;
					
					CGContextSetRGBFillColor(myContext, 0, 1, 0, .5);
					CGContextFillRect(myContext, CGRectMake(doffs, (dheight - (dwidth/2)), dwidth, (dwidth/2)));
				}
				else {
					CGContextSetRGBFillColor(myContext, 0, 0, 1, .5);
					CGContextFillRect(myContext, CGRectMake((float)doffs, 1, (dwidth-1)+partial, 1));
				}
				
				dextra -= partial;
				doffs += partial;
				
				prev_bytes_send = [[stat bytesSend] unsignedLongLongValue];
				prev_bytes_recv = [[stat bytesRecv] unsignedLongLongValue];
				prev_pckts_send = [[stat pcktsSend] unsignedLongLongValue];
				prev_pckts_recv = [[stat pcktsRecv] unsignedLongLongValue];
			}
			
			for (i=0; i < toff; ++i, ++offs, doffs+=dwidth) {
				uint32_t partial = (dextra == 0 ? 0 : (dextra / ((mBuffer-2) - i)));
				
				if (partial * ((mBuffer-2)-i) < dextra)
					partial++;
				
				CGContextSetRGBFillColor(myContext, 0, 0, 1, .5);
				CGContextFillRect(myContext, CGRectMake((float)doffs, 1, (dwidth-1)+partial, 1));
				
				dextra -= partial;
				doffs += partial;
			}
			
		}
		
		/*
		CGContextSetRGBFillColor (myContext, 1, 0, 0, 1);
		CGContextFillRect (myContext, CGRectMake (0, 0, 200, 100 ));
		CGContextSetRGBFillColor (myContext, 0, 0, 1, .5);
		CGContextFillRect (myContext, CGRectMake (0, 0, 100, 200));
		*/
		
		{
			char * statId = (char *)[[NSString stringWithFormat:@"0x%08lx", [mStatId unsignedLongValue]] UTF8String];
			float color[] = {1., 1., 1., mFadeAlpha};
			float size = 9.;
			
			CGContextSelectFont(myContext, "Lucida Grande", size, kCGEncodingMacRoman);
			CGContextSetFillColor(myContext, color);
			CGContextShowTextAtPoint(myContext, 5, [self bounds].size.height - 15, statId, 10);
		}
		
		/*
		if ([self bounds].size.width >= 250)
		{
			uint32_t doffs = [self bounds].size.height - 45;
			float size = 9.;
			float color[] = {1., 1., 1., mFadeAlpha};
			
			for (i = 0, offs=mOffset; i < mBuffer && doffs > 0; ++i, --offs, doffs-=15) {
				if (offs < 0)
					offs = mBuffer - 1;
				
				GlowWormStat * stat = [mSamples objectAtIndex:offs];
				NSString * txt = [NSString stringWithFormat:@"tx b:%llu p:%llu rx b:%llu p:%llu",
					[[stat bytesSend] unsignedLongLongValue], 
					[[stat pcktsSend] unsignedLongLongValue],
					[[stat bytesRecv] unsignedLongLongValue],
					[[stat pcktsRecv] unsignedLongLongValue]];
				
				CGContextSelectFont(myContext, "Lucida Grande", size, kCGEncodingMacRoman);
				CGContextSetFillColor(myContext, color);
				CGContextShowTextAtPoint(myContext, 5, doffs, (char *)[txt UTF8String], [txt length]);
			}
		}
		*/
		
		{
			float size = 9.;
			GlowWormStat * stat = [mSamples objectAtIndex:mOffset];
			
			NSString * bTxStr = [self hrByteCount:[stat bytesSend]];
			NSString * bRxStr = [self hrByteCount:[stat bytesRecv]];
			NSString * pTxStr = [self hrPcktCount:[stat pcktsSend]];
			NSString * pRxStr = [self hrPcktCount:[stat pcktsRecv]];
			
			CGContextSelectFont(myContext, "Lucida Grande", size, kCGEncodingMacRoman);
			
			{
				float color[] = {0., 1., 0., mFadeAlpha};
				CGContextSetFillColor(myContext, color);
				CGContextShowTextAtPoint(myContext, [self bounds].size.width - 67, [self bounds].size.height - 15, (char *)"rx ", 3);
				CGContextShowTextAtPoint(myContext, [self bounds].size.width - 55, [self bounds].size.height - 15, (char *)[bRxStr UTF8String], [bRxStr length]);
				CGContextShowTextAtPoint(myContext, [self bounds].size.width - 55, [self bounds].size.height - 30, (char *)[pRxStr UTF8String], [pRxStr length]);
			}
			
			{
				float color[] = {1., 0., 0., mFadeAlpha};
				CGContextSetFillColor(myContext, color);
				CGContextShowTextAtPoint(myContext, [self bounds].size.width - 67, [self bounds].size.height - 45, (char *)"tx ", 3);
				CGContextShowTextAtPoint(myContext, [self bounds].size.width - 55, [self bounds].size.height - 45, (char *)[bTxStr UTF8String], [bTxStr length]);
				CGContextShowTextAtPoint(myContext, [self bounds].size.width - 55, [self bounds].size.height - 60, (char *)[pTxStr UTF8String], [pTxStr length]);
			}
			
		}
		
		{
//		float dashs[] = {1., 1.};
			float color[] = {.85, .85, .85, mFadeAlpha};
			
//		CGPoint l1 = {{[self bounds].size.width-3., 1.}, {[self bounds].size.width-1., 3}};
//		CGPoint l2 = {{[self bounds].size.width-5., 1.}, {[self bounds].size.width-1., 5}};
			CGPoint l3[] = {{[self bounds].size.width-3., 1.}, {[self bounds].size.width-1., 3}};
			CGPoint l4[] = {{[self bounds].size.width-7., 1.}, {[self bounds].size.width-1., 7}};
			CGPoint l5[] = {{[self bounds].size.width-11., 1.}, {[self bounds].size.width-1., 11}};
			
			//CGContextStrokePath(myContext);
			
			CGContextSetStrokeColor(myContext, color);
			CGContextSetAlpha(myContext, mFadeAlpha);
//		CGContextSetFillColor(myContext, color);
//		CGContextSetLineDash(myContext, 0., dashs, 2);
			CGContextSetLineWidth(myContext, 1);
			CGContextSaveGState(myContext);
			
			CGContextBeginPath(myContext);
			CGContextAddLines(myContext, l3, 2);
//		CGContextDrawPath(myContext, kCGPathEOFillStroke);
			CGContextStrokePath(myContext);
			
			CGContextBeginPath(myContext);
			CGContextAddLines(myContext, l4, 2);
//		CGContextDrawPath(myContext, kCGPathEOFillStroke);
			CGContextStrokePath(myContext);
			
			CGContextBeginPath(myContext);
			CGContextAddLines(myContext, l5, 2);
//		CGContextDrawPath(myContext, kCGPathEOFillStroke);
			CGContextStrokePath(myContext);
			
			CGContextRestoreGState(myContext);
			
			/*
			CGMutablePathRef path = CGPathCreateMutable();
			CGContextBeginPath(myContext);
			CGContextAddLines(myContext, l1, 2);
			*/
			
		}
		
	}
}

- (NSString *)hrByteCount:(NSNumber *)aNum
{
	double num = (double)[aNum unsignedLongLongValue];
	NSString * label = @"";
	
	if (num >= 1099511627776LL) {
		label = @"TB";
		num /= 1099511627776.;
	}
	else if (num >= 1073741824L) {
		label = @"GB";
		num /= 1073741824.;
	}
	else if (num >= 1048576) {
		label = @"MB";
		num /= 1048576.;
	}
	else if (num >= 1024) {
		label = @"KB";
		num /= 1024.;
	}
	
	return [NSString stringWithFormat:@"%.2lf %@", num, label];
}

- (NSString *)hrPcktCount:(NSNumber *)aNum
{
	return [NSString stringWithFormat:@"%llu", [aNum unsignedLongLongValue]];
}

/*
if (prev_bytes_send != 0 && (tmpPt = [[stat bytesSend] unsignedLongLongValue]) > prev_bytes_send) {
	tmpPt -= prev_bytes_send;
	dheight = (uint32_t)(((float)height / (float)mScale) * (float)tmpPt);
	
	 CGRect barRect = {{(float)doffs, (float)(dheight - (dwidth/2.))}, {(float)dwidth, (dwidth/2.)}};
	 CGContextSaveGState(myContext);
	 CGContextBeginPath(myContext);
	 CGContextAddRect(myContext, barRect);
	 CGContextClip(myContext);
	
	begPt.x = (float)doffs;
	begPt.y = 10.;
	endPt.x = (float)doffs + ((float)dwidth-1.) + (float)partial;
	endPt.y = 10.;
	
	shading = CGShadingCreateAxial(getTheRGBColorSpace(), begPt, endPt, axialFunc, 0, 0);
	
	 CGContextDrawShading(myContext, shading);
	 CGShadingRelease( shading );
	 CGContextRestoreGState(myContext);
	
	//GWLog(@"%s.. [%d] tmpPt=%llu, height=%lu, scale=%d, doffs=%lu, dheight=%lu, partial=%lu, dextra=%lu\n", __PRETTY_FUNCTION__, i, tmpPt, height, mScale, doffs, dheight, partial, dextra);
}
else {
	//GWLog(@"%s.. [%d] prev=%llu, tmpPt=%llu", __PRETTY_FUNCTION__, i, prev_bytes_send, tmpPt);
	
	CGContextSetRGBFillColor(myContext, 0, 0, 1, .5);
	CGContextFillRect(myContext, CGRectMake((float)doffs, 1, (dwidth-1)+partial, 1));
}
*/



- (void)printDescription
{
	@synchronized (mSamples)
	{
		int i=0, pos=mOffset+1;
		
		for (; i < mBuffer; ++i, ++pos) {
			GlowWormStat * stat = [mSamples objectAtIndex:pos%mBuffer];
			
			if (stat)
				;
			
			GWLog(@"%s.. [%d : %d] br=%llu, pr=%llu, bs=%llu, ps=%llu", __PRETTY_FUNCTION__, i, (pos%mBuffer),
						[[stat bytesRecv] unsignedLongLongValue],
						[[stat pcktsRecv] unsignedLongLongValue],
						[[stat bytesSend] unsignedLongLongValue],
						[[stat pcktsSend] unsignedLongLongValue]);
		}
	}
}

- (void)randomDataLoop
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	while (TRUE)
	{
		NSAutoreleasePool * pool = [[NSAutoreleasePool alloc] init];
		
		@synchronized (mSamples)
		{
			if (5 != (random() % 10)) {
			//GlowWormStat * prev_stat = [mSamples objectAtIndex:(mOffset==0?mBuffer-1:mOffset-1)];
				GlowWormStat * prev_stat = [mSamples objectAtIndex:mOffset];
				GlowWormStat * next_stat = [[[GlowWormStat alloc] init] autorelease];
				
				uint64_t bytes_send = [[prev_stat bytesSend] unsignedLongLongValue];
				uint64_t bytes_recv = [[prev_stat bytesRecv] unsignedLongLongValue];
				uint64_t pckts_send = [[prev_stat pcktsSend] unsignedLongLongValue];
				uint64_t pckts_recv = [[prev_stat pcktsRecv] unsignedLongLongValue];
				
				bytes_send += random() % 1000;
				bytes_recv += random() % 500;
				pckts_send += random() % 10;
				pckts_recv += random() % 5;
				
				[next_stat setObject:[NSNumber numberWithUnsignedLong:[DataMonitorView getTimeInMicros]] forKey:@"updt_sec"];
				[next_stat setObject:[NSNumber numberWithUnsignedLongLong:bytes_send] forKey:@"bytes_send"];
				[next_stat setObject:[NSNumber numberWithUnsignedLongLong:bytes_recv] forKey:@"bytes_recv"];
				[next_stat setObject:[NSNumber numberWithUnsignedLongLong:pckts_send] forKey:@"pckts_send"];
				[next_stat setObject:[NSNumber numberWithUnsignedLongLong:pckts_recv] forKey:@"pckts_recv"];
				
				[self addSample:next_stat];
//			[self printDescription];
			}
			
		//[self setNeedsDisplay:TRUE];
		}
		
	//sleep(1);
		usleep(mRate/2);
		
		[pool release];
	}
}

- (void)mouseEntered:(NSEvent *)anEvent
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	if (mFadeTimer != nil) {
		[mFadeTimer invalidate];
		[mFadeTimer release];
	}
	
	mFadeTimer = [[NSTimer scheduledTimerWithTimeInterval:0.02 target:self selector:@selector(fadeInStep:) userInfo:nil repeats:YES] retain];
}

- (void)mouseExited:(NSEvent *)anEvent
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	if (mFadeTimer != nil) {
		[mFadeTimer invalidate];
		[mFadeTimer release];
	}
	
	mFadeTimer = [[NSTimer scheduledTimerWithTimeInterval:0.02 target:self selector:@selector(fadeOutStep:) userInfo:nil repeats:YES] retain];
}

- (void)fadeInStep:(NSTimer *)timer
{
	if (mFadeAlpha >= 0.5) {
		[timer invalidate];
		[mFadeTimer release];
		mFadeTimer = nil;
		return;
	}
	
	mFadeAlpha += 0.03;
	[self setNeedsDisplay:TRUE];
}

// - (BOOL)mouse:(NSPoint)aPoint inRect:(NSRect)aRect

- (void)mouseDown:(NSEvent *)anEvent
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	NSRect rect;
	rect.origin.x = [self bounds].size.width-11;
	rect.origin.y = 0;
	rect.size.width = 11;
	rect.size.height = 11;
//NSPoint pt = [NSEvent mouseLocation];
	NSPoint pt = [anEvent locationInWindow];
	
	if (![self mouse:pt inRect:rect]) {
		GWLog(@"%s.. x=%d, y=%d", __PRETTY_FUNCTION__, (int)pt.x, (int)pt.y);
		return;
	}
	
	GWLog(@"%s.. inRect = TRUE", __PRETTY_FUNCTION__);
	
	mLiveResize = TRUE;
	[[self window] setMovableByWindowBackground:FALSE];
}

- (void)mouseDragged:(NSEvent *)anEvent
{
	//GWLog(@"%s", __PRETTY_FUNCTION__);
	
	int dx = [anEvent deltaX];
	int dy = [anEvent deltaY];
	NSRect rect = [[self window] frame];
//NSRect bounds = [[self window] bounds];
	NSRect screen = [[NSScreen mainScreen] visibleFrame];
	
	if (mLiveResize) {
		if (rect.size.width + dx <= 520 && rect.size.width + dx >= 130)
			rect.size.width  += dx;
		
		if (rect.size.height + dy <= 265 && rect.size.height + dy >= 66) {
			rect.size.height += dy;
			rect.origin.y    -= dy;
		}
	}
	else {
		rect.origin.x += dx;
		rect.origin.y -= dy;
		
		if (dx < 0 && rect.origin.x <= 10)
			rect.origin.x = 0;
		else if (dx > 0 && rect.origin.x + rect.size.width >= screen.size.width - 10)
			rect.origin.x = screen.size.width - rect.size.width;
		
		if (dy < 0 && rect.origin.y + rect.size.height >= screen.size.height - 10)
			rect.origin.y = screen.size.height - rect.size.height;
		else if (dy > 0 && rect.origin.y <= 10)
			rect.origin.y = 0;
	}
	
	[[self window] setFrame:rect display:FALSE];
}

- (void)mouseUp:(NSEvent *)anEvent
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
	/*
	if (mLiveResize) {
		CGContextRef myContext = [[NSGraphicsContext currentContext] graphicsPort];
		NSRect bounds = [self bounds];
		
		CGContextSetRGBFillColor(myContext, 1., 1., 1., 1.);
		CGContextFillRect(myContext, CGRectMake(0, 0, bounds.size.width, bounds.size.height));
	}
	*/
	
	mLiveResize = FALSE;
}

- (void)fadeOutStep:(NSTimer *)timer
{
	if (mFadeAlpha <= 0.0) {
		[timer invalidate];
		[mFadeTimer release];
		mFadeTimer = nil;
		return;
	}
	
	mFadeAlpha -= 0.03;
	[self setNeedsDisplay:TRUE];
}

- (void)redrawLoop
{
	GWLog(@"%s", __PRETTY_FUNCTION__);
	
//int offs = mOffset;
	
	while (!mHault)
	{
		usleep(mRate);
		
		@synchronized (mSamples)
		{
			/*
			int prev = (offs == 0 ? mBuffer - 1 : offs - 1);
			
			if (offs == mOffset) {
				[mSamples replaceObjectAtIndex:offs withObject:[mSamples objectAtIndex:prev]];
				mOffset++;
				
				if (mOffset >= mBuffer)
					mOffset = 0;
			}
			
			offs = mOffset;
			*/
			
			[self setNeedsDisplay:TRUE];
		}
	}
}

@end
